package tools.descartes.prisma.agent.wildfly.service;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.UUID;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.jboss.logging.Logger;
import org.jboss.msc.service.Service;
import org.jboss.msc.service.ServiceName;
import org.jboss.msc.service.StartContext;
import org.jboss.msc.service.StartException;
import org.jboss.msc.service.StopContext;
import org.jboss.msc.value.InjectedValue;

import edu.kit.ipd.descartes.mm.applicationlevel.parameterdependencies.ModelVariableCharacterizationType;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyContext;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.BasicComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.CompositeComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Interface;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceProvidingRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceRequiringRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ProvidingDelegationConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RequiringDelegationConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Signature;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.AbstractAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCall;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCallAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.FineGrainedBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ForkAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.InternalAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopIterationCount;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ResourceDemand;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.system.SystemFactory;
import edu.kit.ipd.descartes.mm.deployment.DeploymentFactory;
import edu.kit.ipd.descartes.mm.resourcelandscape.RuntimeEnvironment;
import edu.kit.ipd.descartes.mm.runtimeenvironmentclasses.RuntimeEnvironmentClasses;
import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.impl.CDOModelRepository;
import tools.descartes.prisma.core.util.ModelSkeletonUtil;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;
import tools.descartes.prisma.model.skeleton.SkeletonFactory;

public class ModelExtractionService extends AgentController implements Service<ModelExtractionService> {

	private class CallPaths {
		private Map<String, List<String>> paths = new HashMap<>();

		public void addPath(String interf, String assemblyCtx) {
			List<String> a = paths.get(interf);
			if (a == null) {
				a = new LinkedList<>();
				paths.put(interf, a);
			}
			a.add(assemblyCtx);
		}

		public int getIndex(String interf, String assemblyCtx) {
			List<String> a = paths.get(interf);
			if (a != null) {
				int idx = 0;
				for (String cur : a) {
					if (cur.equals(assemblyCtx)) {
						return idx;
					}
					idx++;
				}
			}
			return -1;
		}
	}

	private static final Logger log = Logger.getLogger(ModelExtractionService.class);

	private InjectedValue<MonitoringService> monitoring = new InjectedValue<>();

	public static final ServiceName SERVICE_NAME = ServiceName.JBOSS.append("modelextraction", "modelextraction");

	private static InetAddress localHost;
	static {
		try {
			localHost = InetAddress.getLocalHost();
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
	}

	public static final ModelRepository repository = new CDOModelRepository("10.0.1.148");

	public static final ModelExtractionService INSTANCE = new ModelExtractionService();

	private ModelSkeleton skeleton;

	private RuntimeEnvironment wildFlyContainer;

	private Map<String, CallPaths> callpaths = new HashMap<>();

	private Queue<AssemblyContext> undeployedAssemblyContexts = new LinkedList<>();

	private Queue<Object> updates = new ConcurrentLinkedQueue<>();

	public ModelExtractionService() {
		super(repository, "application/specj", localHost.getHostName().toString() + "_" + UUID.randomUUID().toString());
	}

	@Override
	public ModelExtractionService getValue() throws IllegalStateException, IllegalArgumentException {
		return this;
	}

	@Override
	public void start(StartContext arg0) throws StartException {
		try {
			localHost = InetAddress.getLocalHost();
		} catch (UnknownHostException e1) {
			e1.printStackTrace();
		}

		skeleton = SkeletonFactory.eINSTANCE.createModelSkeleton();
		skeleton.setRepository(RepositoryFactory.eINSTANCE.createRepository());
		skeleton.setSystem(SystemFactory.eINSTANCE.createSystem());
		skeleton.setDeployment(DeploymentFactory.eINSTANCE.createDeployment());

		try {
			start().get();
			schedule(new ModelUpdater(), 1, TimeUnit.MINUTES);
		} catch (InterruptedException | ExecutionException e) {
			log.error("Error stopping agent.", e);
		}
	}

	@Override
	public void stop(StopContext arg0) {
		try {
			stop().get();
		} catch (InterruptedException | ExecutionException e) {
			log.error("Error stopping agent.", e);
		}
	}

	public void newComponentRecord(ComponentRecord record) {
		updates.add(record);
	}

	public void newInterfaceRecord(InterfaceRecord record) {
		updates.add(record);
	}

	public void newOperationRecord(OperationRecord record) {
		updates.add(record);
	}

	public void newInvocationGroupRecord(InvocationGroupRecord record) {
		updates.add(record);
	}

	@Override
	protected void notifyAddDelegatedElement(EObject newElement, Map<String, String> additionalProperties) {
		if (newElement instanceof RuntimeEnvironment) {
			RuntimeEnvironment env = (RuntimeEnvironment) newElement;
			if (env.getOfClass() == RuntimeEnvironmentClasses.OS_VM
					&& localHost.getHostAddress().equals(additionalProperties.get("container.ip"))) {
				wildFlyContainer = env;
			}
		}
	}

	private class ModelUpdater implements Runnable {

		@Override
		public void run() {
			try {
				Object current;

				boolean changed = false;

				while ((current = updates.poll()) != null) {
					if (current instanceof InvocationGroupRecord) {
						handleInvocationGroupRecord((InvocationGroupRecord) current);
						changed = true;
					}
				}

				if (changed) {
					apply(skeleton);
				}
			} catch (Throwable ex) {
				log.error("Error updating skeleton.", ex);
			}
		}

		private void handleInvocationGroupRecord(InvocationGroupRecord record) {
			InvocationGroupRecord caller = record.getOperation().getInterface().getComponent().getCaller();
			if (caller != null) {
				// We have information on the incoming calls as well as the
				// outgoing calls.
				addComponent(caller, record);
			} else {
				// no information about outgoing calls.
				addComponent(record, null);
			}
		}
	}

	private void addComponent(InvocationGroupRecord incomingGrec, InvocationGroupRecord outgoingGrec) {
		ComponentRecord incomingCrec = incomingGrec.getOperation().getInterface().getComponent();
		String incomingInterfaceName = incomingGrec.getOperation().getInterface().getInterfaceName();
		ComponentRecord outgoingCrec = null;
		String outgoingInterfaceName = null;
		if (outgoingGrec != null) {
			outgoingCrec = outgoingGrec.getOperation().getInterface().getComponent();
			outgoingInterfaceName = outgoingGrec.getOperation().getInterface().getInterfaceName();
		}

		// 1. add the incoming interface to the repository
		Interface incomingInterface = RepositoryFactory.eINSTANCE.createInterface();
		incomingInterface.setName(incomingInterfaceName);
		incomingInterface = ModelSkeletonUtil.insert(skeleton.getRepository().getInterfaces(), incomingInterface);

		Signature incomingSignature = RepositoryFactory.eINSTANCE.createSignature();
		incomingSignature.setName(getSignatureName(incomingGrec));
		incomingSignature = ModelSkeletonUtil.insert(incomingInterface.getSignatures(), incomingSignature);

		// 2. add the outgoing interface to the repository (if applicable)
		Interface outgoingInterface = null;
		Signature outgoingSignature = null;
		if (outgoingGrec != null) {
			outgoingInterface = RepositoryFactory.eINSTANCE.createInterface();
			outgoingInterface.setName(outgoingInterfaceName);
			outgoingInterface = ModelSkeletonUtil.insert(skeleton.getRepository().getInterfaces(), outgoingInterface);
			outgoingSignature = RepositoryFactory.eINSTANCE.createSignature();
			outgoingSignature.setName(getSignatureName(outgoingGrec));
			outgoingSignature = ModelSkeletonUtil.insert(outgoingInterface.getSignatures(), outgoingSignature);
		}

		// 3. Create the implementation component for the caller component
		BasicComponent implementationComponent = RepositoryFactory.eINSTANCE.createBasicComponent();
		implementationComponent.setName(incomingCrec.getComponentClass());
		implementationComponent = (BasicComponent) ModelSkeletonUtil.insert(skeleton.getRepository().getComponents(),
				implementationComponent);
		String providingPortName = getInterfaceRoleName(incomingCrec.getComponentClass(), incomingInterface.getName(),
				false);
		InterfaceProvidingRole callerProvidingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
		callerProvidingRole.setName(providingPortName);
		callerProvidingRole.setInterface(incomingInterface);
		callerProvidingRole = ModelSkeletonUtil.insert(implementationComponent.getInterfaceProvidingRoles(),
				callerProvidingRole);
		InterfaceRequiringRole callerRequiringRole = null;
		int callerRoleIdx = 0;
		if (outgoingInterface != null) {
			// Here we check whether there are external calls for the same
			// interface
			// but to different target assembly contexts. If yes, we need to
			// distinguish the two requiring ports
			callerRoleIdx = getRequiringPortIndex(incomingCrec, outgoingGrec.getOperation().getInterface());
			String roleName = getInterfaceRoleName(incomingCrec.getComponentClass(), outgoingInterface.getName(), true);
			if (callerRoleIdx > 0) {
				roleName = roleName + ":" + callerRoleIdx;
			}
			callerRequiringRole = RepositoryFactory.eINSTANCE.createInterfaceRequiringRole();
			callerRequiringRole.setName(roleName);
			callerRequiringRole.setInterface(outgoingInterface);
			callerRequiringRole = ModelSkeletonUtil.insert(implementationComponent.getInterfaceRequiringRoles(),
					callerRequiringRole);
		}
		FineGrainedBehavior behavior = createComponentInternalBehavior(implementationComponent, callerProvidingRole,
				incomingSignature);
		if (callerRequiringRole != null) {
			addExternalCall(incomingCrec.getComponentClass(), implementationComponent, behavior, callerRequiringRole,
					outgoingSignature);
		}
		RepositoryComponent callerComponent = implementationComponent;

		// 4. Now add the called component (if applicable)
		InterfaceProvidingRole calledProvidingRole = null;
		RepositoryComponent calledComponent = null;
		boolean isRemoteCall = false;
		if (callerRequiringRole != null) {
			isRemoteCall = isRemoteCall(incomingCrec, outgoingCrec);
			calledComponent = createCalledComponent(outgoingCrec, isRemoteCall);
			String calledComponentName;
			if (calledComponent instanceof CompositeComponent) {
				calledComponentName = outgoingCrec.getComponentName();
			} else {
				calledComponentName = outgoingCrec.getComponentClass();
			}
			String calledProvingRoleName = getInterfaceRoleName(calledComponentName, outgoingInterfaceName, false);
			calledProvidingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
			calledProvidingRole.setName(calledProvingRoleName);
			calledProvidingRole.setInterface(outgoingInterface);
			calledProvidingRole = ModelSkeletonUtil.insert(calledComponent.getInterfaceProvidingRoles(),
					calledProvidingRole);
		}

		// 5. Now wrap the implementation in a composite component (if
		// necessary)
		CompositeComponent compositeComponent = createCompositeComponent(incomingCrec, callerComponent);
		if (compositeComponent != null) {

			// Create an assembly context for embedding the component and
			// connect its interface providing role with a role on the composite
			// component
			AssemblyContext assembly = RepositoryFactory.eINSTANCE.createAssemblyContext();
			assembly.setName(incomingCrec.getComponentName());
			assembly.setEncapsulatedComponent(implementationComponent);
			assembly = ModelSkeletonUtil.insert(compositeComponent.getAssemblyContexts(), assembly);

			String providingRemotePortName = getInterfaceRoleName(incomingCrec.getComponentName(),
					incomingInterfaceName, false);
			InterfaceProvidingRole remoteProvidingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
			remoteProvidingRole.setName(providingRemotePortName);
			remoteProvidingRole.setInterface(incomingInterface);
			remoteProvidingRole = ModelSkeletonUtil.insert(compositeComponent.getInterfaceProvidingRoles(),
					remoteProvidingRole);

			ProvidingDelegationConnector providingConnector = RepositoryFactory.eINSTANCE
					.createProvidingDelegationConnector();
			providingConnector.setOuterInterfaceProvidingRole(remoteProvidingRole);
			providingConnector.setInnerInterfaceProvidingRole(callerProvidingRole);
			providingConnector.setAssemblyContext(assembly);
			ModelSkeletonUtil.insert(compositeComponent.getProvidingDelegationConnectors(), providingConnector);

			InterfaceRequiringRole remoteRequiringRole = null;
			if (callerRequiringRole != null) {
				if (isRemoteCall) {
					// Only add a interface requiring role to the composite
					// component if it is a remote call, i.e., a call to another
					// composite
					// component.
					String requiringRemotePortName = getInterfaceRoleName(incomingCrec.getComponentName(),
							outgoingInterfaceName, true);
					if (callerRoleIdx > 0) {
						requiringRemotePortName = requiringRemotePortName + ":" + callerRoleIdx;
					}
					remoteRequiringRole = RepositoryFactory.eINSTANCE.createInterfaceRequiringRole();
					remoteRequiringRole.setName(requiringRemotePortName);
					remoteRequiringRole.setInterface(outgoingInterface);
					remoteRequiringRole = ModelSkeletonUtil.insert(compositeComponent.getInterfaceRequiringRoles(),
							remoteRequiringRole);

					RequiringDelegationConnector requiringConnector = RepositoryFactory.eINSTANCE
							.createRequiringDelegationConnector();
					requiringConnector.setOuterInterfaceRequiringRole(remoteRequiringRole);
					requiringConnector.setInnerInterfaceRequiringRole(callerRequiringRole);
					requiringConnector.setAssemblyContext(assembly);
					ModelSkeletonUtil.insert(compositeComponent.getRequiringDelegationConnectors(), requiringConnector);
				} else {
					// Local call: Also add an assembly context for the called
					// component within the composite component
					// and connect it with an assembly connector.
					AssemblyContext calledAssembly = RepositoryFactory.eINSTANCE.createAssemblyContext();
					calledAssembly.setName(outgoingCrec.getComponentName());
					calledAssembly.setEncapsulatedComponent(calledComponent);
					calledAssembly = ModelSkeletonUtil.insert(compositeComponent.getAssemblyContexts(), calledAssembly);

					AssemblyConnector connector = RepositoryFactory.eINSTANCE.createAssemblyConnector();
					connector.setInterfaceProvidingRole(calledProvidingRole);
					connector.setInterfaceRequiringRole(callerRequiringRole);
					connector.setProvidingAssemblyContext(calledAssembly);
					connector.setRequiringAssemblyContext(assembly);
					ModelSkeletonUtil.insert(compositeComponent.getAssemblyConnectors(), connector);
				}
			}
			callerProvidingRole = remoteProvidingRole;
			callerRequiringRole = remoteRequiringRole;
			callerComponent = compositeComponent;
		}

		// 6. Now create the assembly contexts
		AssemblyContext callerContext = RepositoryFactory.eINSTANCE.createAssemblyContext();
		callerContext.setName(getAssemblyContextName(callerComponent, incomingCrec));
		callerContext.setEncapsulatedComponent(callerComponent);
		callerContext = ModelSkeletonUtil.insert(skeleton.getSystem().getAssemblyContexts(), callerContext);
		if (callerRequiringRole != null) {
			AssemblyContext calledContext = RepositoryFactory.eINSTANCE.createAssemblyContext();
			calledContext.setName(getAssemblyContextName(calledComponent, outgoingCrec));
			calledContext.setEncapsulatedComponent(calledComponent);
			calledContext = ModelSkeletonUtil.insert(skeleton.getSystem().getAssemblyContexts(), calledContext);

			AssemblyConnector connector = RepositoryFactory.eINSTANCE.createAssemblyConnector();
			connector.setInterfaceProvidingRole(calledProvidingRole);
			connector.setInterfaceRequiringRole(callerRequiringRole);
			connector.setProvidingAssemblyContext(calledContext);
			connector.setRequiringAssemblyContext(callerContext);
			ModelSkeletonUtil.insert(skeleton.getSystem().getAssemblyConnectors(), connector);
		}

		// 7. Expose the provided remote interfaces as system interfaces (if
		// this port is marked as system interface)
		String systemProvidingRoleName = localHost.getHostAddress() + "/" + incomingInterfaceName;
		// Do not create a new providing/requiring role for the system, only use
		// the existing providing/requiring role.
		// This way a user can control which interfaces are exposed at the
		// system level by creating them manually.
		InterfaceProvidingRole systemProvidingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
		systemProvidingRole.setName(systemProvidingRoleName);
		systemProvidingRole.setInterface(incomingInterface);
		systemProvidingRole = ModelSkeletonUtil.find(
				getContainerScope().getSystemScope().getSystem().getInterfaceProvidingRoles(), systemProvidingRole);
		if (systemProvidingRole != null) {
			skeleton.getSystem().getInterfaceProvidingRoles()
					.add(ModelSkeletonUtil.adopt(getModelRepository(), systemProvidingRole));

			ProvidingDelegationConnector systemProvidingConnector = RepositoryFactory.eINSTANCE
					.createProvidingDelegationConnector();
			systemProvidingConnector.setOuterInterfaceProvidingRole(systemProvidingRole);
			systemProvidingConnector.setInnerInterfaceProvidingRole(callerProvidingRole);
			systemProvidingConnector.setAssemblyContext(callerContext);
			ModelSkeletonUtil.insert(skeleton.getSystem().getProvidingDelegationConnectors(), systemProvidingConnector);
		}

		if (callerRequiringRole != null) {
			String systemRequiringRoleName = getInterfaceRoleName(incomingCrec.getComponentName(),
					outgoingInterfaceName, true);
			InterfaceRequiringRole systemRequiringRole = RepositoryFactory.eINSTANCE.createInterfaceRequiringRole();
			systemRequiringRole.setName(systemRequiringRoleName);
			systemRequiringRole.setInterface(outgoingInterface);
			systemRequiringRole = ModelSkeletonUtil.find(
					getContainerScope().getSystemScope().getSystem().getInterfaceRequiringRoles(), systemRequiringRole);
			if (systemRequiringRole != null) {
				skeleton.getSystem().getInterfaceRequiringRoles()
						.add(ModelSkeletonUtil.adopt(getModelRepository(), systemRequiringRole));

				RequiringDelegationConnector systemRequiringConnector = RepositoryFactory.eINSTANCE
						.createRequiringDelegationConnector();
				systemRequiringConnector.setInnerInterfaceRequiringRole(callerRequiringRole);
				systemRequiringConnector.setOuterInterfaceRequiringRole(systemRequiringRole);
				systemRequiringConnector.setAssemblyContext(callerContext);
				ModelSkeletonUtil.insert(skeleton.getSystem().getRequiringDelegationConnectors(),
						systemRequiringConnector);
			}
		}

		// 8. Remember to deploy this context (only for the caller component)
		undeployedAssemblyContexts.add(callerContext);
	}

	private FineGrainedBehavior createComponentInternalBehavior(BasicComponent implementationComponent,
			InterfaceProvidingRole implementationProvidingRole, Signature signature) {
		FineGrainedBehavior behavior = ServicebehaviorFactory.eINSTANCE.createFineGrainedBehavior();
		behavior.setDescribedSignature(signature);
		behavior.setInterfaceProvidingRole(implementationProvidingRole);
		behavior = ModelSkeletonUtil.insert(implementationComponent.getFineGrainedBehavior(), behavior);
		if (behavior.getBehavior() == null) {
			behavior.setBehavior(ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior());
		}
		if (!isJMSComponent(implementationComponent.getName())) {
			if (behavior.getBehavior().getActions().isEmpty()) {
				InternalAction internalAction = ServicebehaviorFactory.eINSTANCE.createInternalAction();
				ResourceDemand cpuDemand = ServicebehaviorFactory.eINSTANCE.createResourceDemand();
				cpuDemand.setName("CPU");
				cpuDemand.setCharacterization(ModelVariableCharacterizationType.EMPIRICAL);
				// TODO: how to cope with reosurce types!!!
				cpuDemand.setResourceType(null);
				internalAction.getResourceDemand().add(cpuDemand);
				ModelSkeletonUtil.insert(behavior.getBehavior().getActions(), internalAction);
			}
		}
		return behavior;
	}

	private String getAssemblyContextName(RepositoryComponent component, ComponentRecord crec) {
		if (component instanceof CompositeComponent) {
			// Composite components are named after the component name in the
			// component record.
			// We also use that as the assembly context name.
			return component.getName();
		} else {
			// Basic components are named after the component class in the
			// component record.
			// We use the component name instead for the assembly context since
			// it is used for resolving
			// remote calls.
			return crec.getComponentName();
		}
	}

	private void addExternalCall(String componentName, BasicComponent implementationComponent,
			FineGrainedBehavior behavior, InterfaceRequiringRole requiringRole, Signature outgoingSignature) {
		ExternalCallAction callAction = ServicebehaviorFactory.eINSTANCE.createExternalCallAction();
		ExternalCall call = ServicebehaviorFactory.eINSTANCE.createExternalCall();
		call.setName(requiringRole.getName() + "#" + outgoingSignature.getName());
		call.setSignature(outgoingSignature);
		call.setInterfaceRequiringRole(requiringRole);

		if (isJMSComponent(componentName)) {
			if (behavior.getBehavior().getActions().isEmpty()) {
				ForkAction fork = ServicebehaviorFactory.eINSTANCE.createForkAction();
				fork.setWithSynchronizationBarrier(false);
				fork.getForkedBehaviors().get(0).getActions().add(callAction);
				behavior.getBehavior().getActions().add(fork);
			}
		} else {
			LoopAction loop = ServicebehaviorFactory.eINSTANCE.createLoopAction();
			LoopIterationCount iterationCount = ServicebehaviorFactory.eINSTANCE.createLoopIterationCount();
			iterationCount.setName(call.getName());
			iterationCount.setCharacterization(ModelVariableCharacterizationType.EMPIRICAL);
			loop.setLoopIterationCount(iterationCount);
			loop.setLoopBodyBehavior(ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior());
			loop.getLoopBodyBehavior().getActions().add(callAction);

			// We assume that the external calls are inserted in
			// alphabetical order.
			EList<AbstractAction> actions = behavior.getBehavior().getActions();
			for (int i = 0; i < actions.size(); i++) {
				if (actions.get(i) instanceof LoopAction) {
					LoopAction curLoop = (LoopAction) actions.get(i);
					int compareResult = curLoop.getLoopIterationCount().getName().compareTo(call.getName());
					if (compareResult == 0) {
						return;
					} else if (compareResult > 0) {
						actions.add(i, loop);
						return;
					}
				}
			}
		}
	}

	private CompositeComponent createCompositeComponent(ComponentRecord crec,
			RepositoryComponent implementationComponent) {
		String compositeName = getCompositeComponentName(crec);
		if (compositeName != null) {
			CompositeComponent compositeComponent = RepositoryFactory.eINSTANCE.createCompositeComponent();
			compositeComponent.setName(compositeName);
			return (CompositeComponent) ModelSkeletonUtil.insert(skeleton.getRepository().getComponents(),
					compositeComponent);
		}
		return null;
	}

	private RepositoryComponent createCalledComponent(ComponentRecord callee,
			boolean remoteCall) {
		// check whether this call goes to a component in the same composite or
		// not
		if (remoteCall) {
			String calleeCompositeName = getCompositeComponentName(callee);
			if (calleeCompositeName != null) {
				CompositeComponent compositeComponent = RepositoryFactory.eINSTANCE.createCompositeComponent();
				compositeComponent.setName(calleeCompositeName);
				return (CompositeComponent) ModelSkeletonUtil.insert(skeleton.getRepository().getComponents(),
						compositeComponent);
			}
		}
		if (callee.getComponentClass() != null) {
			BasicComponent basicComponent = RepositoryFactory.eINSTANCE.createBasicComponent();
			basicComponent.setName(callee.getComponentClass());
			return basicComponent;
		}
		throw new IllegalStateException("Could not create called component for " + callee.getComponentName());
	}

	private boolean isRemoteCall(ComponentRecord caller, ComponentRecord callee) {
		String callerCompositeName = getCompositeComponentName(caller);
		String calleeCompositeName = getCompositeComponentName(callee);
		// A call is considered local if the two components are in the same
		// composite
		return (calleeCompositeName == null) || (callerCompositeName == null)
				|| !calleeCompositeName.equals(callerCompositeName);
	}

	private boolean isJMSComponent(String componentName) {
		return "javax.jms.Queue".equals(componentName) || "javax.jms.Topic".equals(componentName);
	}

	private String getCompositeComponentName(ComponentRecord crec) {
		String componentPath = crec.getComponentName();
		if (componentPath.startsWith("/")) {
			componentPath = componentPath.substring(1);
		}
		if (componentPath.endsWith("/")) {
			componentPath = componentPath.substring(0, componentPath.length() - 1);
		}
		int idx = componentPath.lastIndexOf("/");
		if (idx < 0) {
			return null;
		} else {
			return componentPath.substring(0, idx);
		}
	}

	private int getRequiringPortIndex(ComponentRecord incomingCrec, InterfaceRecord outgoingIrec) {
		CallPaths paths = callpaths.get(incomingCrec.getComponentName());
		if (paths == null) {
			paths = new CallPaths();
			callpaths.put(incomingCrec.getComponentName(), paths);
		}
		paths.addPath(outgoingIrec.getInterfaceName(), outgoingIrec.getComponent().getComponentName());
		int idx = paths.getIndex(outgoingIrec.getInterfaceName(), outgoingIrec.getComponent().getComponentName());
		if (idx < 0) {
			throw new IllegalStateException();
		}
		return idx;
	}

	public InjectedValue<MonitoringService> getMonitoringService() {
		return monitoring;
	}

	private String getSignatureName(InvocationGroupRecord record) {
		String name = record.getOperation().getInterface().getInterfaceName() + "#"
				+ record.getOperation().getOperationName();
		if (!record.getName().isEmpty()) {
			name += "_" + record.getName();
		}
		return name;
	}

	private String getInterfaceRoleName(String component, String interf, boolean req) {
		return component + "#" + (req ? "Req:" : "Prov:") + interf;
	}
}
